@@ -10,32 +10,19 @@ API itself. Maybe use flickr.activity.userPhotos
 (https://www.flickr.com/services/api/flickr.activity.userPhotos.html)
 when group_id=SELF.
 """
-
-__author__ = ['Kyle Mahan <kyle@kylewm.com>']
-
 import copy
-import datetime
-import functools
-import itertools
-import json
 import logging
 import requests
-import source
-import sys
-import mf2py
 import mf2util
-import urllib2
-import urlparse
+import urllib.parse
 
-import appengine_config
 from oauth_dropins.webutil import util
 from oauth_dropins import flickr_auth
 
-from apiclient.errors import HttpError
-from apiclient.http import BatchHttpRequest
-
+from . import source
 
 class Flickr(source.Source):
+  """Flickr source class. See file docstring and Source class for details."""
 
   DOMAIN = 'flickr.com'
   BASE_URL = 'https://www.flickr.com/'
@@ -64,11 +51,11 @@ class Flickr(source.Source):
     self._user_id = user_id
     self._path_alias = path_alias
 
-  def call_api_method(self, method, params={}):
+  def call_api_method(self, method, params=None):
     """Call a Flickr API method.
     """
     return flickr_auth.call_api_method(
-      method, params, self.access_token_key, self.access_token_secret)
+      method, params or {}, self.access_token_key, self.access_token_secret)
 
   def upload(self, params, file):
     """Upload a photo or video via the Flickr API.
@@ -76,12 +63,13 @@ class Flickr(source.Source):
     return flickr_auth.upload(
       params, file, self.access_token_key, self.access_token_secret)
 
-  def create(self, obj, include_link=False, ignore_formatting=False):
+  def create(self, obj, include_link=source.OMIT_LINK,
+             ignore_formatting=False):
     """Creates a photo, comment, or favorite.
 
     Args:
       obj: ActivityStreams object
-      include_link: boolean
+      include_link: string
       ignore_formatting: boolean
 
     Returns:
@@ -93,12 +81,14 @@ class Flickr(source.Source):
     return self._create(obj, preview=False, include_link=include_link,
                         ignore_formatting=ignore_formatting)
 
-  def preview_create(self, obj, include_link=False, ignore_formatting=False):
+  def preview_create(self, obj, include_link=source.OMIT_LINK,
+                     ignore_formatting=False):
     """Preview creation of a photo, comment, or favorite.
 
     Args:
       obj: ActivityStreams object
-      include_link: boolean
+      include_link: string
+      ignore_formatting: boolean
 
     Returns:
       a CreationResult whose description will be an HTML summary of
@@ -108,7 +98,8 @@ class Flickr(source.Source):
     return self._create(obj, preview=True, include_link=include_link,
                         ignore_formatting=ignore_formatting)
 
-  def _create(self, obj, preview, include_link=False, ignore_formatting=False):
+  def _create(self, obj, preview, include_link=source.OMIT_LINK,
+              ignore_formatting=False):
     """Creates or previews creating for the previous two methods.
 
     https://www.flickr.com/services/api/upload.api.html
@@ -119,7 +110,8 @@ class Flickr(source.Source):
     Args:
       obj: ActivityStreams object
       preview: boolean
-      include_link: boolean
+      include_link: string
+      ignore_formatting: boolean
 
     Return:
       a CreationResult
@@ -149,7 +141,7 @@ class Flickr(source.Source):
         content = None
 
       # add original post link
-      if include_link:
+      if include_link == source.INCLUDE_LINK:
         content = ((content + '\n\n') if content else '') + link_text
 
       if preview:
@@ -180,13 +172,21 @@ class Flickr(source.Source):
       if name:
         params.append(('title', name))
       if content:
-        params.append(('description', content))
+        params.append(('description', content.encode('utf-8')))
       if hashtags:
-        params.append(
-          ('tags', ','.join('"%s"' % t if ' ' in t else t for t in hashtags)))
+        params.append(('tags', ','.join(('"%s"' % t if ' ' in t else t)
+                                        for t in hashtags)))
 
       file = util.urlopen(video_url or image_url)
-      resp = self.upload(params, file)
+      try:
+        resp = self.upload(params, file)
+      except requests.exceptions.ConnectionError as e:
+        if str(e.args[0]).startswith('Request exceeds 10 MiB limit'):
+          msg = 'Sorry, photos and videos must be under 10MB.'
+          return source.creation_result(error_plain=msg, error_html=msg)
+        else:
+          raise
+
       photo_id = resp.get('id')
       resp.update({
         'type': 'post',
@@ -228,7 +228,7 @@ class Flickr(source.Source):
           'link to a Flickr photo or to an original post that publishes a '
           '<a href="http://indiewebcamp.com/rel-syndication">rel-syndication</a> link to Flickr.')
 
-      if include_link:
+      if include_link == source.INCLUDE_LINK:
         content += '\n\n' + link_text
       if preview:
         return source.creation_result(
@@ -237,7 +237,7 @@ class Flickr(source.Source):
 
       resp = self.call_api_method('flickr.photos.comments.addComment', {
         'photo_id': base_id,
-        'comment_text': content,
+        'comment_text': content.encode('utf-8'),
       })
       resp = resp.get('comment', {})
       resp.update({
@@ -299,7 +299,34 @@ class Flickr(source.Source):
           tag = copy.copy(tag)
           tag['id'] = id
           people[id] = tag
-    return people.values()
+    return list(people.values())
+
+  def delete(self, id):
+    """Deletes a photo. The authenticated user must have created it.
+
+    Args:
+      id: int or string, photo id to delete
+
+    Returns: CreationResult, content is Flickr API response dict
+    """
+    # delete API call has no response
+    self.call_api_method('flickr.photos.delete', {'photo_id': id})
+    return source.creation_result({
+      'type': 'delete',
+      'url': self.photo_url(self.user_id(), id),
+    })
+
+  def preview_delete(self, id):
+    """Previews deleting a photo.
+
+    Args:
+      id: int or string, photo id to delete
+
+    Returns: CreationResult
+    """
+    return source.creation_result(
+      description='<span class="verb">delete</span> <a href="%s">this photo</a>.' %
+        self.photo_url(self.user_id(), id))
 
   def get_activities_response(self, user_id=None, group_id=None, app_id=None,
                               activity_id=None, start_index=0, count=0,
@@ -421,26 +448,27 @@ class Flickr(source.Source):
     profile_url = person.get('profileurl', {}).get('_content')
     if profile_url:
       try:
-        resp = util.urlopen(profile_url)
-        profile_json = mf2py.parse(doc=resp, url=profile_url)
-        # personal site is likely the first non-flickr url
+        profile_json = util.fetch_mf2(url=profile_url)
         urls = profile_json.get('rels', {}).get('me', [])
-        obj['urls'] = [{'value': u} for u in urls]
-        obj['url'] = next(
-          (u for u in urls if not u.startswith('https://www.flickr.com/')),
-          None)
-      except urllib2.URLError, e:
+        if urls:
+          obj['url'] = urls[0]
+        if len(urls) > 1:
+          obj['urls'] = [{'value': u} for u in urls]
+      except requests.RequestException as e:
+        util.interpret_http_exception(e)
         logging.warning('could not fetch user homepage %s', profile_url)
 
     return self.postprocess_object(obj)
 
-  def get_comment(self, comment_id, activity_id, activity_author_id=None):
+  def get_comment(self, comment_id, activity_id, activity_author_id=None,
+                  activity=None):
     """Returns an ActivityStreams comment object.
 
     Args:
       comment_id: string comment id
       activity_id: string activity id, required
       activity_author_id: string activity author id, ignored
+      activity: activity object (optional)
     """
     resp = self.call_api_method('flickr.photos.comments.getList', {
       'photo_id': activity_id,
@@ -449,7 +477,7 @@ class Flickr(source.Source):
       logging.debug('checking comment id %s', comment.get('id'))
       # comment id is the in form ###-postid-commentid
       if (comment.get('id') == comment_id or
-              comment.get('id').split('-')[-1] == comment_id):
+          comment.get('id').split('-')[-1] == comment_id):
         logging.debug('found comment matching %s', comment_id)
         return self.comment_to_object(comment, activity_id)
 
@@ -497,7 +525,7 @@ class Flickr(source.Source):
         'url': photo_permalink,
         'id': self.tag_uri(photo.get('id')),
         'image': {
-          'url': u'https://farm{}.staticflickr.com/{}/{}_{}_{}.jpg'.format(
+          'url': 'https://farm{}.staticflickr.com/{}/{}_{}_{}.jpg'.format(
             photo.get('farm'), photo.get('server'),
             photo.get('id'), photo.get('secret'), 'b'),
         },
@@ -515,11 +543,20 @@ class Flickr(source.Source):
     }
 
     if isinstance(owner, dict):
+      username = owner.get('username')
+      # Flickr API is evidently inconsistent in what it puts in realname vs
+      # username vs path_alias. if username has spaces, it's probably actually
+      # real name, so use path alias instead.
+      # https://github.com/snarfed/bridgy/issues/687
+      # https://www.flickr.com/groups/51035612836@N01/discuss/72157625945600254
+      if not username or ' ' in username:
+        username = owner.get('path_alias')
+
       activity['object']['author'] = {
         'objectType': 'person',
-        'displayName': owner.get('realname') or owner.get('username'),
-        'username': owner.get('username'),
-        'id': self.tag_uri(owner.get('username')),
+        'displayName': owner.get('realname') or username,
+        'username': username,
+        'id': self.tag_uri(username),
         'image': {
           'url': self.get_user_image(owner.get('iconfarm'),
                                      owner.get('iconserver'),
@@ -531,14 +568,14 @@ class Flickr(source.Source):
       activity['object']['tags'] = [{
           'objectType': 'hashtag',
           'id': self.tag_uri(tag.get('id')),
-          'url': u'https://www.flickr.com/search?tags={}'.format(
+          'url': 'https://www.flickr.com/search?tags={}'.format(
             tag.get('_content')),
           'displayName': tag.get('raw'),
         } for tag in photo.get('tags', {}).get('tag', [])]
-    elif isinstance(photo.get('tags'), basestring):
+    elif isinstance(photo.get('tags'), str):
       activity['object']['tags'] = [{
         'objectType': 'hashtag',
-        'url': u'https://www.flickr.com/search?tags={}'.format(
+        'url': 'https://www.flickr.com/search?tags={}'.format(
           tag.strip()),
         'displayName': tag.strip(),
       } for tag in photo.get('tags').split(' ') if tag.strip()]
@@ -582,10 +619,10 @@ class Flickr(source.Source):
         },
       },
       'created': util.maybe_timestamp_to_rfc3339(photo_activity.get('favedate')),
-      'url': u'{}#liked-by-{}'.format(
+      'url': '{}#liked-by-{}'.format(
         photo_activity.get('url'), person.get('nsid')),
       'object': {'url': photo_activity.get('url')},
-      'id': self.tag_uri(u'{}_liked_by_{}'.format(
+      'id': self.tag_uri('{}_liked_by_{}'.format(
         photo_activity.get('flickr_id'), person.get('nsid'))),
       'objectType': 'activity',
       'verb': 'like',
@@ -632,8 +669,8 @@ class Flickr(source.Source):
     ref: https://www.flickr.com/services/api/misc.buddyicons.html
     """
     if server == 0:
-      return u'https://www.flickr.com/images/buddyicon.gif'
-    return u'https://farm{}.staticflickr.com/{}/buddyicons/{}.jpg'.format(
+      return 'https://www.flickr.com/images/buddyicon.gif'
+    return 'https://farm{}.staticflickr.com/{}/buddyicons/{}.jpg'.format(
       farm, server, author)
 
   def user_id(self):
@@ -686,13 +723,13 @@ class Flickr(source.Source):
     Returns:
       string, the photo URL
     """
-    return u'https://www.flickr.com/photos/%s/%s/' % (user_id, photo_id)
+    return 'https://www.flickr.com/photos/%s/%s/' % (user_id, photo_id)
 
   @classmethod
-  def post_id(cls, url):
+  def base_id(cls, url):
     """Used when publishing comments or favorites. Flickr photo ID is the
     3rd path component rather than the first.
     """
-    parts = urlparse.urlparse(url).path.split('/')
+    parts = urllib.parse.urlparse(url).path.split('/')
     if len(parts) >= 4 and parts[1] == 'photos':
       return parts[3]
