@@ -6,24 +6,19 @@ http://help.instagram.com/448523408565555
 https://groups.google.com/forum/m/#!topic/instagram-api-developers/DAO7OriVFsw
 https://groups.google.com/forum/#!searchin/instagram-api-developers/private
 """
-
-__author__ = ['Ryan Barrett <granary@ryanb.org>']
-
 import datetime
-import json
 import logging
 import operator
 import re
 import string
-import urllib
-import urllib2
-import urlparse
+import urllib.parse, urllib.request
 import xml.sax.saxutils
 
-import appengine_config
 from oauth_dropins.webutil import util
+from oauth_dropins.webutil.util import json_dumps, json_loads
 import requests
-import source
+
+from . import source
 
 # Maps Instagram media type to ActivityStreams objectType.
 OBJECT_TYPES = {'image': 'photo', 'video': 'video'}
@@ -39,16 +34,38 @@ API_MEDIA_POPULAR_URL = 'https://api.instagram.com/v1/media/popular'
 API_MEDIA_LIKES_URL = 'https://api.instagram.com/v1/media/%s/likes'
 API_COMMENT_URL = 'https://api.instagram.com/v1/media/%s/comments'
 
-HTML_MEDIA = 'https://www.instagram.com/p/%s/'
+HTML_BASE_URL = util.read('instagram_scrape_base') or 'https://www.instagram.com/'
+HTML_MEDIA = HTML_BASE_URL + 'p/%s/'
+HTML_PROFILE = HTML_BASE_URL + '%s/'
+HTML_PRELOAD_RE = re.compile(
+  r'^/graphql/query/\?query_hash=[^&]*&(amp;)?variables=(%7B%7D|{})$')
+# the query hash here comes (i think) from inside a .js file served by IG, so
+# we'd have to fetch and scrape that to get it dynamically. not worth it yet.
+HTML_LIKES_URL = HTML_BASE_URL + 'graphql/query/?query_hash=d5d763b1e2acf209d62d22d184488e57&variables={"shortcode":"%s","include_reel":false,"first":100}'
+HTML_DATA_RE = re.compile(r"""
+  <script\ type="text/javascript">
+  window\.(_sharedData\ =|__additionalDataLoaded\('[^']+',)\ *
+  (.+?)
+  \)?;</script>""", re.VERBOSE)
 
 # URL-safe base64 encoding. used in Instagram.id_to_shortcode()
 BASE64 = string.ascii_uppercase + string.ascii_lowercase + string.digits + '-_'
 
 MENTION_RE = re.compile(r'@([A-Za-z0-9._]+)')
 
+# global lock for backing off scraping due to rate limiting.
+RATE_LIMIT_BACKOFF = datetime.timedelta(seconds=5 * 60)
+_last_rate_limited = None      # datetime
+_last_rate_limited_exc = None  # requests.HTTPError
+
+AUTO_ALT_TEXT_PREFIXES = (
+  'No photo description available.',
+  'Image may contain: ',
+)
+
 
 class Instagram(source.Source):
-  """Implements the ActivityStreams API for Instagram."""
+  """Instagram source class. See file docstring and Source class for details."""
 
   DOMAIN = 'instagram.com'
   BASE_URL = 'https://www.instagram.com/'
@@ -63,7 +80,8 @@ class Instagram(source.Source):
   </blockquote>
   """
 
-  def __init__(self, access_token=None, allow_comment_creation=False, scrape=False):
+  def __init__(self, access_token=None, allow_comment_creation=False,
+               scrape=False, cookie=None):
     """Constructor.
 
     If an OAuth access token is provided, it will be passed on to Instagram.
@@ -75,40 +93,51 @@ class Instagram(source.Source):
       allow_comment_creation: boolean, optionally disable comment creation,
         useful if the app is not approved to create comments.
       scrape: boolean, whether to scrape instagram.com's HTML (True) or use
-      the API (False)
+        the API (False)
+      cookie: string, optional sessionid cookie to use when scraping.
     """
     self.access_token = access_token
     self.allow_comment_creation = allow_comment_creation
     self.scrape = scrape
+    self.cookie = cookie
 
   def urlopen(self, url, **kwargs):
-    """Wraps urllib2.urlopen() and passes through the access token."""
-    log_url = url
+    """Wraps :func:`urllib2.urlopen()` and passes through the access token."""
     if self.access_token:
       # TODO add access_token to the data parameter for POST requests
       url = util.add_query_params(url, [('access_token', self.access_token)])
-    resp = util.urlopen(urllib2.Request(url, **kwargs))
-    return resp if kwargs.get('data') else json.loads(resp.read()).get('data')
+    resp = util.urlopen(urllib.request.Request(url, **kwargs))
+    return (resp if kwargs.get('data')
+            else source.load_json(resp.read(), url).get('data'))
 
   @classmethod
   def user_url(cls, username):
-    return '%s%s/' % (cls.BASE_URL, username)
+    if username:
+      return '%s%s/' % (cls.BASE_URL, username)
 
-  def get_actor(self, user_id=None):
+  @classmethod
+  def media_url(cls, shortcode):
+    if shortcode:
+      return '%sp/%s/' % (cls.BASE_URL, shortcode)
+
+  def get_actor(self, user_id=None, **kwargs):
     """Returns a user as a JSON ActivityStreams actor dict.
 
     Args:
       user_id: string id or username. Defaults to 'self', ie the current user.
-
-    Raises: InstagramAPIError
+      kwargs: if scraping, passed through to get_activities_response(). Raises
+        AssertionError if provided and not scraping.
     """
     if user_id is None:
-      assert self.scrape is False, 'get_actor() requires user_id when scraping'
+      if self.scrape:
+        return {}
       user_id = 'self'
 
     if self.scrape:
-      return self.get_activities_response(group_id=source.SELF, user_id=user_id
-                                         ).get('actor')
+      resp = self.get_activities_response(
+        group_id=source.SELF, user_id=user_id, **kwargs)
+      items = resp.get('items')
+      return items[0].get('actor') if items else {}
     else:
       return self.user_to_actor(util.trim_nulls(
         self.urlopen(API_USER_URL % user_id) or {}))
@@ -119,7 +148,8 @@ class Instagram(source.Source):
                               fetch_replies=False, fetch_likes=False,
                               fetch_shares=False, fetch_events=False,
                               fetch_mentions=False, search_query=None,
-                              scrape=False, cookie=None, **kwargs):
+                              scrape=False, cookie=None, ignore_rate_limit=False,
+                              **kwargs):
     """Fetches posts and converts them to ActivityStreams activities.
 
     See method docstring in source.py for details. app_id is ignored.
@@ -144,35 +174,62 @@ class Instagram(source.Source):
     May populate a custom 'ig_like_count' property in media objects. (Currently
     only when scraping.)
 
-    Args (beyond Source.get_activities_response):
+    Args:
       scrape: if True, scrapes HTML from instagram.com instead of using the API.
         Populates the user's actor object in the 'actor' response field.
         Useful for apps that haven't yet been approved in the new permissions
-        approval process. Currently only supports group_id=SELF.
+        approval process. Currently only supports group_id=SELF. Also supports
+        passing a shortcode as activity_id as well as the internal API id.
         http://developers.instagram.com/post/133424514006/instagram-platform-update
       cookie: string, only used if scrape=True
-
-    Raises: InstagramAPIError
+      ignore_rate_limit: boolean, for scraping, always make an HTTP request,
+        even if we've been rate limited recently
+      **: see :meth:`Source.get_activities_response`
     """
+    if group_id is None:
+      group_id = source.FRIENDS
+
     if scrape or self.scrape:
+      cookie = cookie or self.cookie
       if not (activity_id or
               (group_id == source.SELF and user_id) or
               (group_id == source.FRIENDS and cookie)):
         raise NotImplementedError(
           'Scraping only supports activity_id, user_id and group_id=@self, or cookie and group_id=@friends.')
-      return self._scrape(user_id=user_id, activity_id=activity_id, cookie=cookie,
-                          fetch_extras=fetch_replies or fetch_likes, cache=cache)
+      elif fetch_likes and not cookie:
+        raise NotImplementedError('Scraping likes requires a cookie.')
+
+      # cache rate limited responses and short circuit
+      global _last_rate_limited, _last_rate_limited_exc
+      now = datetime.datetime.now()
+      if not ignore_rate_limit and _last_rate_limited:
+        retry = _last_rate_limited + RATE_LIMIT_BACKOFF
+        if now < retry:
+          logging.info('Remembered rate limit at %s, waiting until %s to try again.',
+                       _last_rate_limited, retry)
+          assert _last_rate_limited_exc
+          raise _last_rate_limited_exc
+
+      try:
+        return self._scrape(
+          user_id=user_id, group_id=group_id, activity_id=activity_id, count=count,
+          cookie=cookie, fetch_extras=fetch_replies or fetch_likes, cache=cache)
+      except Exception as e:
+        code, body = util.interpret_http_exception(e)
+        if not ignore_rate_limit and code in ('401', '429', '503'):
+          logging.info('Got rate limited! Remembering for %s', str(RATE_LIMIT_BACKOFF))
+          _last_rate_limited = now
+          _last_rate_limited_exc = e
+        raise
 
     if user_id is None:
       user_id = 'self'
-    if group_id is None:
-      group_id = source.FRIENDS
 
     if search_query:
       if search_query.startswith('#'):
         search_query = search_query[1:]
       else:
-        raise NotImplementedError(
+        raise ValueError(
           'Instagram only supports search over hashtags, so search_query must '
           'begin with the # character.')
 
@@ -205,53 +262,70 @@ class Instagram(source.Source):
           activities += [self.like_to_object(user, l['id'], l['link'])
                          for l in liked]
 
-    except urllib2.HTTPError, e:
+    except urllib.error.HTTPError as e:
       code, body = util.interpret_http_exception(e)
       # instagram api should give us back a json block describing the
       # error. but if it's an error for some other reason, it probably won't
       # be properly formatted json.
       try:
-        body_obj = json.loads(body) if body else {}
+        body_obj = json_loads(body) if body else {}
       except ValueError:
         body_obj = {}
 
       if body_obj.get('meta', {}).get('error_type') == 'APINotFoundError':
-        logging.exception(body_obj.get('meta', {}).get('error_message'))
+        logging.warning(body_obj.get('meta', {}).get('error_message'), stack_info=True)
       else:
         raise e
 
     return self.make_activities_base_response(activities)
 
-  def _scrape(self, user_id=None, activity_id=None, cookie=None,
-              fetch_extras=False, cache=None):
+  def _scrape(self, user_id=None, group_id=None, activity_id=None, cookie=None,
+              count=None, fetch_extras=False, cache=None, shortcode=None):
     """Scrapes a user's profile or feed and converts the media to activities.
 
     Args:
       user_id: string
-      activity_id: string
+      activity_id: string, e.g. '1020355224898358984_654594'
+      count: integer, number of activities to fetch and return, None for all
       fetch_extras: boolean
       cookie: string
+      shortcode: string, e.g. '4pB6vEx87I'
 
-    Returns: list of activities
+    Returns:
+      dict activities API response
     """
-    assert user_id or activity_id or cookie
+    cookie = cookie or self.cookie
+    assert user_id or activity_id or shortcode or cookie
+    assert not (activity_id and shortcode)
 
-    url = (HTML_MEDIA % self.id_to_shortcode(activity_id) if activity_id
-           else self.user_url(user_id) if user_id
-           else self.BASE_URL)
-    kwargs = {}
+    if not shortcode:
+      shortcode = self.id_to_shortcode(activity_id)
+
+    url = (HTML_MEDIA % shortcode if shortcode
+           else HTML_PROFILE % user_id if user_id and group_id == source.SELF
+           else HTML_BASE_URL)
+    get_kwargs = {'allow_redirects': False}
     if cookie:
-      kwargs = {'headers': {'Cookie': cookie}}
-    resp = util.requests_get(url, allow_redirects=False, **kwargs)
+      if not cookie.startswith('sessionid='):
+        cookie = 'sessionid=' + cookie
+      get_kwargs['headers'] = {'Cookie': cookie}
+
+    resp = util.requests_get(url, **get_kwargs)
     if ((cookie and 'not-logged-in' in resp.text) or
-        (resp.status_code in ('301', '302') and
+        (resp.status_code in (301, 302) and
          '/accounts/login' in resp.headers.get('Location', ''))):
-      resp.status_code = '401'
+      resp.status_code = 401
       raise requests.HTTPError('401 Unauthorized', response=resp)
+    elif resp.status_code == 404:
+      if activity_id:
+        return self._scrape(shortcode=activity_id, cookie=cookie, count=count)
+      # otherwise not found, fall through and return empty response
+    else:
+      resp.raise_for_status()
 
-    activities, actor = self.html_to_activities(resp.text)
+    activities, actor = self.html_to_activities(resp.text, cookie=cookie, count=count)
 
-    if fetch_extras and not activity_id:
+    if fetch_extras:
       # batch get cached counts of comments and likes for all activities
       cached = {}
       # don't update the cache until the end, in case we hit an error before
@@ -273,8 +347,14 @@ class Instagram(source.Source):
 
         if (likes and likes != cached.get(likes_key) or
             comments and comments != cached.get(comments_key)):
+          if not activity_id and not shortcode:
+            url = activity['url'].replace(self.BASE_URL, HTML_BASE_URL)
+            resp = util.requests_get(url, **get_kwargs)
+            resp.raise_for_status()
+          # otherwise resp is a fetch of just this activity; reuse it
+
           full_activity, _ = self.html_to_activities(
-            util.requests_get(activity['url']).text)
+            resp.text, cookie=cookie, count=count, fetch_extras=fetch_extras)
           if full_activity:
             activities[i] = full_activity[0]
             cache_updates.update({likes_key: likes, comments_key: comments})
@@ -286,54 +366,62 @@ class Instagram(source.Source):
     resp['actor'] = actor
     return resp
 
-  def get_comment(self, comment_id, activity_id=None, activity_author_id=None):
+  def get_comment(self, comment_id, activity_id=None, activity_author_id=None,
+                  activity=None):
     """Returns an ActivityStreams comment object.
 
     Args:
       comment_id: string comment id
       activity_id: string activity id, required
       activity_author_id: string activity author id. Ignored.
+      activity: activity object, optional. Avoids fetching the activity.
     """
-    activities = self.get_activities(activity_id=activity_id)
-    if activities:
+    if not activity:
+      activity = self._get_activity(None, activity_id)
+    if activity:
       tag_id = self.tag_uri(comment_id)
-      for reply in activities[0].get('object', {}).get('replies', {}).get('items', []):
+      for reply in activity.get('object', {}).get('replies', {}).get('items', []):
         if reply.get('id') == tag_id:
           return reply
 
-  def get_share(self, activity_user_id, activity_id, share_id):
+  def get_share(self, activity_user_id, activity_id, share_id, activity=None):
     """Not implemented. Returns None. Resharing isn't a feature of Instagram.
     """
     return None
 
-  def create(self, obj, include_link=False, ignore_formatting=False):
+  def create(self, obj, include_link=source.OMIT_LINK, ignore_formatting=False):
     """Creates a new comment or like.
 
     Args:
       obj: ActivityStreams object
-      include_link: boolean
+      include_link: string
       ignore_formatting: boolean
 
-    Returns: a CreationResult. if successful, content will have and 'id' and
+    Returns:
+      a CreationResult. if successful, content will have and 'id' and
              'url' keys for the newly created Instagram object
     """
     return self._create(obj, include_link=include_link, preview=False,
                         ignore_formatting=ignore_formatting)
 
-  def preview_create(self, obj, include_link=False, ignore_formatting=False):
+  def preview_create(self, obj, include_link=source.OMIT_LINK,
+                     ignore_formatting=False):
     """Preview a new comment or like.
 
     Args:
       obj: ActivityStreams object
-      include_link: boolean
+      include_link: string
+      ignore_formatting: boolean
 
-    Returns: a CreationResult. if successful, content and description
+    Returns:
+      a CreationResult. if successful, content and description
              will describe the new instagram object.
     """
     return self._create(obj, include_link=include_link, preview=True,
                         ignore_formatting=ignore_formatting)
 
-  def _create(self, obj, include_link=False, preview=None, ignore_formatting=False):
+  def _create(self, obj, include_link=source.OMIT_LINK, preview=None,
+              ignore_formatting=False):
     """Creates a new comment or like.
 
     The OAuth access token must have been created with scope=comments+likes (or
@@ -348,10 +436,11 @@ class Instagram(source.Source):
 
     Args:
       obj: ActivityStreams object
-      include_link: boolean
+      include_link: string
       preview: boolean
 
-    Returns: a CreationResult. if successful, content will have and 'id' and
+    Returns:
+      a CreationResult. if successful, content will have and 'id' and
              'url' keys for the newly created Instagram object
     """
     # TODO: validation, error handling
@@ -373,15 +462,14 @@ class Instagram(source.Source):
           abort=True,
           error_plain='Cannot publish comments on Instagram',
           error_html='<a href="http://instagram.com/developer/endpoints/comments/#post_media_comments">Cannot publish comments</a> on Instagram. The Instagram API technically supports creating comments, but <a href="http://stackoverflow.com/a/26889101/682648">anecdotal</a> <a href="http://stackoverflow.com/a/20229275/682648">evidence</a> suggests they are very selective about which applications they approve to do so.')
-      content = obj.get('content', '').encode('utf-8')
-      content = content if ignore_formatting else source.html_to_text(content)
+      content = self._content_for_create(obj)
       if preview:
         return source.creation_result(
           content=content,
           description='<span class="verb">comment</span> on <a href="%s">'
                       'this post</a>:\n%s' % (base_url, self.embed_post(base_obj)))
 
-      self.urlopen(API_COMMENT_URL % base_id, data=urllib.urlencode({
+      self.urlopen(API_COMMENT_URL % base_id, data=urllib.parse.urlencode({
         'access_token': self.access_token,
         'text': content,
       }))
@@ -415,7 +503,7 @@ class Instagram(source.Source):
       logging.info('posting like for media id id=%s, url=%s',
                    base_id, base_url)
       # no response other than success/failure
-      self.urlopen(API_MEDIA_LIKES_URL % base_id, data=urllib.urlencode({
+      self.urlopen(API_MEDIA_LIKES_URL % base_id, data=urllib.parse.urlencode({
         'access_token': self.access_token
       }))
       # TODO use the stored user_json rather than looking it up each time.
@@ -482,6 +570,7 @@ class Instagram(source.Source):
       }],
       'attachments': [{
         'objectType': 'video' if 'videos' in media else 'image',
+        'url': media.get('link'),
         # ActivityStreams 2.0 allows image to be a JSON array.
         # http://jasnell.github.io/w3c-socialwg-activitystreams/activitystreams2.html#link
         'image': sorted(
@@ -509,22 +598,23 @@ class Instagram(source.Source):
         'displayName': tag,
         # TODO: url
       } for tag in media.get('tags', [])] +
-      [self.user_to_actor(user.get('user'))
-       for user in media.get('users_in_photo', [])] +
-      [self.like_to_object(user, id, media.get('link'))
-       for user in media.get('likes', {}).get('data', [])] +
-      [{
-        'objectType': 'person',
-        'id': self.tag_uri(mention.group(1)),
-        'displayName': mention.group(1),
-        'url': self.user_url(mention.group(1)),
-        'startIndex': mention.start(),
-        'length': mention.end() - mention.start(),
-        } for mention in MENTION_RE.finditer(content)],
+      [self.user_to_actor(u.get('user'))
+       for u in media.get('users_in_photo', [])] +
+      [self.like_to_object(u, id, media.get('link'))
+       for u in media.get('likes', {}).get('data', [])] +
+      self._mention_tags_from_content(content)
     }
 
+    # alt text
+    # https://instagram-press.com/blog/2018/11/28/creating-a-more-accessible-instagram/
+    alt = media.get('accessibility_caption')
+    if alt and not any(alt.startswith(prefix) for prefix in AUTO_ALT_TEXT_PREFIXES):
+      for att in object['attachments']:
+        for img in att.get('image', []):
+          img['displayName'] = alt
+
     for version in ('standard_resolution', 'low_resolution', 'thumbnail'):
-      image = media.get('images').get(version)
+      image = media.get('images', {}).get(version)
       if image:
         object['image'] = {'url': image.get('url')}
         break
@@ -539,7 +629,7 @@ class Instagram(source.Source):
     if 'location' in media:
       media_loc = media.get('location', {})
       object['location'] = {
-        'id': media_loc.get('id'),
+        'id': self.tag_uri(media_loc.get('id')),
         'displayName': media_loc.get('name'),
         'latitude': media_loc.get('point', {}).get('latitude'),
         'longitude': media_loc.get('point', {}).get('longitude'),
@@ -551,6 +641,16 @@ class Instagram(source.Source):
 
     return self.postprocess_object(object)
 
+  def _mention_tags_from_content(self, content):
+    return [{
+      'objectType': 'person',
+      'id': self.tag_uri(mention.group(1)),
+      'displayName': mention.group(1),
+      'url': self.user_url(mention.group(1)),
+      'startIndex': mention.start(),
+      'length': mention.end() - mention.start(),
+    } for mention in MENTION_RE.finditer(content)]
+
   def comment_to_object(self, comment, media_id, media_url):
     """Converts a comment to an object.
 
@@ -562,6 +662,7 @@ class Instagram(source.Source):
     Returns:
       an ActivityStreams object dict, ready to be JSON-encoded
     """
+    content = comment.get('text') or ''
     return self.postprocess_object({
       'objectType': 'comment',
       'id': self.tag_uri(comment.get('id')),
@@ -569,9 +670,10 @@ class Instagram(source.Source):
       'url': '%s#comment-%s' % (media_url, comment.get('id')) if media_url else None,
       # TODO: add PST time zone
       'published': util.maybe_timestamp_to_rfc3339(comment.get('created_time')),
-      'content': comment.get('text'),
+      'content': content,
       'author': self.user_to_actor(comment.get('from')),
       'to': [{'objectType': 'group', 'alias': '@public'}],
+      'tags': self._mention_tags_from_content(content),
     })
 
   def like_to_object(self, liker, media_id, media_url):
@@ -616,14 +718,19 @@ class Instagram(source.Source):
     if not id or not username:
       return actor
 
-    urls = sum((util.extract_links(user.get(field)) for field in ('website', 'bio')),
-               [])
-    if urls:
-      actor['url'] = urls[0]
-      if len(urls) > 1:
-        actor['urls'] = [{'value': u} for u in urls]
-    else:
-      actor['url'] = self.user_url(username)
+    urls = [self.user_url(username)] + sum(
+      (util.extract_links(user.get(field)) for field in ('website', 'bio')), [])
+    actor.update({
+      'url': urls[0],
+      'urls': [{'value': u} for u in urls] if len(urls) > 1 else None
+    })
+
+    private = user.get('is_private')
+    if private is not None:
+      actor['to'] = [{
+        'objectType': 'group',
+        'alias': '@private' if private else '@public',
+      }]
 
     actor.update({
       'displayName': user.get('full_name') or username,
@@ -660,10 +767,12 @@ class Instagram(source.Source):
     if not id:
       return None
 
-    if isinstance(id, basestring):
-      id = int(id.split('_')[0])
+    if isinstance(id, str):
+      parts = id.split('_')
+      if not util.is_int(parts[0]):
+        return id
+      id = int(parts[0])
 
-    A = ord('A')
     chars = []
     while id > 0:
       id, rem = divmod(id, 64)
@@ -671,104 +780,233 @@ class Instagram(source.Source):
 
     return ''.join(reversed(chars))
 
-  def html_to_activities(self, html):
+  def html_to_activities(self, html, cookie=None, count=None, fetch_extras=False):
     """Converts Instagram HTML to ActivityStreams activities.
 
     The input HTML may be from:
+
     * a user's feed, eg https://www.instagram.com/ while logged in
     * a user's profile, eg https://www.instagram.com/snarfed/
     * a photo or video, eg https://www.instagram.com/p/BBWCSrfFZAk/
 
     Args:
       html: unicode string
+      cookie: string, optional sessionid cookie to be used for subsequent HTTP
+        fetches, if necessary.
+      count: integer, number of activities to return, None for all
+      fetch_extras: whether to make extra HTTP fetches to get likes, etc.
 
     Returns:
-      ([ActivityStreams activities], ActivityStreams viewer actor)
+      tuple, ([ActivityStreams activities], ActivityStreams viewer actor)
     """
+    cookie = cookie or self.cookie
     # extract JSON data blob
-    script_start = '<script type="text/javascript">window._sharedData = '
-    start = html.find(script_start)
-    if start == -1:
+    # (can also get just this JSON by adding ?__a=1 to any IG URL.)
+    matches = HTML_DATA_RE.findall(html)
+    if not matches:
+      # Instagram sometimes returns 200 with incomplete HTML. often it stops at
+      # the end of one of the <style> tags inside <head>. not sure why.
+      logging.warning('JSON script tag not found!')
       return [], None
 
-    start += len(script_start)
-    end = html.index(';</script>', start)
-    data = json.loads(html[start:end])
-
-    entry_data = data.get('entry_data', {})
-    activities = []
-
     # find media
     medias = []
     profile_user = None
-    for page in entry_data.get('FeedPage', []):
-      medias.extend(page.get('feed', {}).get('media', {}).get('nodes', []))
-    for page in entry_data.get('ProfilePage', []):
-      profile_user = page.get('user', {})
-      medias.extend(profile_user.get('media', {}).get('nodes', []))
-    medias.extend(page.get('media') for page in entry_data.get('PostPage', []))
 
+    for match in matches:
+      data = util.trim_nulls(json_loads(match[1]))
+      entry_data = data.get('entry_data', {})
+
+      # home page ie news feed
+      for page in entry_data.get('FeedPage', []):
+        edges = page.get('graphql', {}).get('user', {})\
+                    .get('edge_web_feed_timeline', {}).get('edges', [])
+        medias.extend(e.get('node') for e in edges
+                      if e.get('node', {}).get('__typename') not in
+                      ('GraphSuggestedUserFeedUnit',))
+
+      if 'user' in data:
+        edges = data['user'].get('edge_web_feed_timeline', {}).get('edges', [])
+        medias.extend(e.get('node') for e in edges)
+
+      # profiles
+      for page in entry_data.get('ProfilePage', []):
+        profile_user = page.get('graphql', {}).get('user', {})
+        medias.extend(edge['node'] for edge in
+          profile_user.get('edge_owner_to_timeline_media', {}).get('edges', [])
+          if edge.get('node'))
+
+      # individual photo/video permalinks
+      for page in [data] + entry_data.get('PostPage', []):
+        media = page.get('graphql', {}).get('shortcode_media')
+        if media:
+          medias.append(media)
+
+    if not medias:
+      # As of 2018-02-15, embedded JSON in logged in https://www.instagram.com/
+      # no longer has any useful data. Need to do a second header link fetch.
+      soup = util.parse_html(html)
+      link = soup.find('link', href=HTML_PRELOAD_RE)
+      if link:
+        url = urllib.parse.urljoin(HTML_BASE_URL, link['href'])
+        data = self._scrape_json(url, cookie=cookie)
+        edges = data.get('data', {}).get('user', {})\
+                    .get('edge_web_feed_timeline', {}).get('edges', [])
+        medias = [e.get('node') for e in edges]
+
+    if count:
+      medias = medias[:count]
+
+    activities = []
     for media in util.trim_nulls(medias):
-      # preprocess to make its field names match the API's
-      dims = media.get('dimensions', {})
-      owner = media.get('owner', {})
+      shortcode = media.get('code') or media.get('shortcode')
+      likes = media.get('edge_media_preview_like') or {}
+      if shortcode and fetch_extras and likes.get('count') and not likes.get('edges'):
+        # extra GraphQL fetch to get likes, as of 8/2018
+        data = self._scrape_json(HTML_LIKES_URL % shortcode, cookie=cookie)
+        likes['edges'] = data.get('data', {}).get('shortcode_media', {})\
+                             .get('edge_liked_by', {}).get('edges', [])
+      activities.append(self._json_media_node_to_activity(media, user=profile_user))
+
+    actor = None
+    user = self._json_user_to_user(data.get('config', {}).get('viewer') or profile_user)
+    if user:
+      actor = self.user_to_actor(user)
+
+    return activities, actor
+
+  @staticmethod
+  def _scrape_json(url, cookie=None):
+    """Fetches and returns JSON from www.instagram.com."""
+    headers = {}
+    if cookie:
+      if not cookie.startswith('sessionid='):
+        cookie = 'sessionid=' + cookie
+      headers = {'Cookie': cookie}
+
+    resp = util.requests_get(url, allow_redirects=False, headers=headers)
+    resp.raise_for_status()
+
+    try:
+      return json_loads(resp.text)
+    except ValueError as e:
+      msg = "Couldn't decode response as JSON:\n%s" % resp.text
+      logging.error(msg, stack_info=True)
+      resp.status_code = 504
+      raise requests.HTTPError('504 Bad response from Instagram\n' + msg,
+                               response=resp)
+
+  def _json_media_node_to_activity(self, media, user=None):
+    """Converts Instagram HTML JSON media node to ActivityStreams activity.
+
+    Args:
+      media: dict, subset of Instagram HTML JSON representing a single photo
+        or video
+      user: top-level user object from Instagram HTML JSON, e.g. on a profile page
+
+    Returns:
+      dict, ActivityStreams activity
+    """
+    # preprocess to make its field names match the API's
+    owner = media.get('owner', {})
+    owner_id = owner.get('id')
+    if user and user.get('id') == owner_id:
+      owner.update(user)
+
+    dims = media.get('dimensions', {})
+    image_url = media.get('display_src') or media.get('display_url') or ''
+    link = self.media_url(media.get('code') or media.get('shortcode'))
+    media.update({
+      'link': link,
+      'user': self._json_user_to_user(owner),
+      'created_time': media.get('date') or media.get('taken_at_timestamp'),
+      'caption': {'text': media.get('edge_media_to_caption', {})
+                               .get('edges', [{}])[0].get('node', {}).get('text', '')},
+      'images': {'standard_resolution': {
+        'url': image_url.replace('\/', '/'),
+        'width': dims.get('width'),
+        'height': dims.get('height'),
+      }},
+      'users_in_photo': (media.get('usertags', {}).get('nodes', []) +
+                         [e.get('node', {}) for e in
+                          media.get('edge_media_to_tagged_user', {}).get('edges', [])]),
+    })
+
+    id = media.get('id')
+    owner_id = owner.get('id')
+    if id and owner_id:
+      media['id'] = '%s_%s' % (id, owner_id)
+
+    comments = (media.get('comments') or media.get('edge_media_to_comment') or
+                media.get('edge_media_to_parent_comment') or {})
+    media['comments'] = {
+      'data': [c.get('node') for c in comments.get('edges', [])],
+      'count': comments.get('count'),
+    }
+
+    likes = media.get('likes') or media.get('edge_media_preview_like') or {}
+    media['likes'] = {
+      'data': [l.get('node') for l in likes.get('edges', [])],
+      'count': likes.get('count'),
+    }
+
+    for obj in [media] + media['comments']['data'] + media['likes']['data']:
+      obj.setdefault('user', obj.get('owner') or {})
+      user = obj['user'] or obj
+      user['profile_picture'] = user.get('profile_pic_url', '').replace('\/', '/')
+
+    for c in media['comments']['data']:
+      c['from'] = c['user']
+      c['created_time'] = c['created_at']
+
+    if media.get('is_video'):
       media.update({
-        'link': 'https://www.instagram.com/p/%s/' % media.get('code'),
-        'user': owner,
-        'created_time': media.get('date'),
-        'caption': {'text': media.get('caption')},
-        'images': {'standard_resolution': {
-          'url': media.get('display_src').replace('\/', '/'),
+        'type': 'video',
+        'videos': {'standard_resolution': {
+          'url': media.get('video_url', '').replace('\/', '/'),
           'width': dims.get('width'),
           'height': dims.get('height'),
         }},
-        'users_in_photo': media.get('usertags', {}).get('nodes', []),
       })
 
-      id = media.get('id')
-      owner_id = owner.get('id')
-      if id and owner_id:
-        media['id'] = '%s_%s' % (id, owner_id)
+    activity = self.media_to_activity(util.trim_nulls(media))
+    obj = activity['object']
+    obj['ig_like_count'] = media['likes'].get('count', 0)
 
-      comments = media.setdefault('comments', {}).setdefault('nodes', [])
-      likes = media.setdefault('likes', {}).setdefault('nodes', [])
-      for obj in [media] + comments + likes:
-        obj['user']['profile_picture'] = \
-          obj['user'].get('profile_pic_url', '').replace('\/', '/')
+    # multi-photo
+    children = media.get('edge_sidecar_to_children', {}).get('edges', [])
+    if children:
+      obj['attachments'] = []
+      for child in children:
+        child_activity = self._json_media_node_to_activity(child.get('node'))
+        for att in child_activity['object']['attachments']:
+          if not att.get('url'):
+            att['url'] = link
+          obj['attachments'].append(att)
 
-      media['comments']['data'] = comments
-      for c in media['comments']['data']:
-        c['from'] = c['user']
-        c['created_time'] = c['created_at']
-      media['likes']['data'] = [l['user'] for l in likes]
+    self.postprocess_object(obj)
+    return super(Instagram, self).postprocess_activity(activity)
 
-      if media.get('is_video'):
-        media.update({
-          'type': 'video',
-          'videos': {'standard_resolution': {
-            'url': media.get('video_url', '').replace('\/', '/'),
-            'width': dims.get('width'),
-            'height': dims.get('height'),
-          }},
-        })
+  def _json_user_to_user(self, user):
+    """Converts an Instagram HTML JSON user to an API actor.
 
-      activity = self.media_to_activity(util.trim_nulls(media))
-      activity['object']['ig_like_count'] = media['likes'].get('count', 0)
-      self.postprocess_object(activity['object'])
-      activities.append(super(Instagram, self).postprocess_activity(activity))
+    Args:
+      media: dict, HTML JSON user
 
-    actor = None
-    viewer = data.get('config', {}).get('viewer') or profile_user or None
-    if viewer:
-      profile = viewer.get('profile_pic_url')
-      if profile:
-        viewer['profile_picture'] = profile.replace('\/', '/')
-      website = viewer.get('external_url')
-      if website:
-        viewer['website'] = website.replace('\/', '/')
-      viewer.setdefault('bio', viewer.get('biography'))
-      actor = self.user_to_actor(viewer)
-      if viewer.get('is_private'):
-        actor['to'] = [{'objectType':'group', 'alias':'@private'}]
+    Returns:
+      dict, API user object
+    """
+    if not user:
+      return None
 
-    return activities, actor
+    if user.get('user'):
+      user = user['user']
+    profile = user.get('profile_pic_url')
+    if profile:
+      user['profile_picture'] = profile.replace('\/', '/')
+    website = user.get('external_url')
+    if website:
+      user['website'] = website.replace('\/', '/')
+    user.setdefault('bio', user.get('biography'))
+
+    return user
