@@ -9,22 +9,18 @@ activity's privacy settings. It's set to a group with alias @public or @private,
 or unset if unknown.
 http://activitystrea.ms/specs/json/targeting/1.0/#anchor3
 """
-
-__author__ = ['Ryan Barrett <granary@ryanb.org>']
-
 import collections
 import copy
+from html import unescape
 import logging
-import mimetypes
 import re
-import urlparse
-import html2text
+import urllib.parse
 
+import brevity
 from bs4 import BeautifulSoup
-import requests
-
-import appengine_config
+import html2text
 from oauth_dropins.webutil import util
+from oauth_dropins.webutil.util import json_dumps, json_loads
 
 ME = '@me'
 SELF = '@self'
@@ -32,13 +28,31 @@ ALL = '@all'
 FRIENDS = '@friends'
 APP = '@app'
 SEARCH = '@search'
+BLOCKS = '@blocks'
 
-RSVP_TO_EVENT = {
-  'rsvp-yes': 'attending',
-  'rsvp-no': 'notAttending',
-  'rsvp-maybe': 'maybeAttending',
-  'invite': 'invited',
-  }
+# values for create's include_link param
+OMIT_LINK = 'omit'
+INCLUDE_LINK = 'include'
+INCLUDE_IF_TRUNCATED = 'if truncated'
+
+# maps each AS1 RSVP verb to the collection inside an event object that the
+# RSVPing actor would go into.
+RSVP_VERB_TO_COLLECTION = collections.OrderedDict((  # in priority order
+  ('rsvp-yes', 'attending'),
+  ('rsvp-no', 'notAttending'),
+  ('rsvp-maybe', 'maybeAttending'),
+  ('rsvp-interested', 'interested'),
+  ('invite', 'invited'),
+))
+VERBS_WITH_OBJECT = {
+  'follow',
+  'like',
+  'react',
+  'repost',
+  'share',
+} | set(RSVP_VERB_TO_COLLECTION.keys())
+
+HTML_ENTITY_RE = re.compile(r'&#?[a-zA-Z0-9]+;')
 
 # maps lower case string short name to Source subclass. populated by SourceMeta.
 sources = {}
@@ -46,37 +60,71 @@ sources = {}
 CreationResult = collections.namedtuple('CreationResult', [
   'content', 'description', 'abort', 'error_plain', 'error_html'])
 
+class RateLimited(BaseException):
+  """Raised when an API rate limits us, and we may have a partial result.
 
-def strip_html_tags(str):
-  """Returns the text content of an HTML string, with tags removed."""
-  return BeautifulSoup(str, 'html.parser').get_text('')
+  Attributes:
+    partial: the partial result, if any. Usually a list.
+  """
+  def __init__(self, *args, **kwargs):
+    self.partial = kwargs.pop('partial', None)
+    super(RateLimited, self).__init__(*args, **kwargs)
 
 
-def html_to_text(html):
-  """Converts string html to string text with html2text."""
+def html_to_text(html, **kwargs):
+  """Converts string html to string text with html2text.
+
+  Args:
+    **kwargs: html2text options
+      https://github.com/Alir3z4/html2text/blob/master/docs/usage.md#available-options
+  """
   if html:
     h = html2text.HTML2Text()
     h.unicode_snob = True
     h.body_width = 0  # don't wrap lines
     h.ignore_links = True
     h.ignore_images = True
+    for key, val in kwargs.items():
+      setattr(h, key, val)
+
+    # hacky monkey patch fix for html2text escaping sequences that are
+    # significant in markdown syntax. the X\\Y replacement depends on knowledge
+    # of html2text's internals, specifically that it replaces RE_MD_*_MATCHER
+    # with \1\\\2. :(:(:(
+    html2text.config.RE_MD_DOT_MATCHER = \
+      html2text.config.RE_MD_PLUS_MATCHER = \
+      html2text.config.RE_MD_DASH_MATCHER = \
+        re.compile(r'(X)\\(Y)')
+
     return '\n'.join(
       # strip trailing whitespace that html2text adds to ends of some lines
-      line.rstrip() for line in h.unescape(h.handle(html)).splitlines())
+      line.rstrip() for line in unescape(h.handle(html)).splitlines())
+
+
+def load_json(body, url):
+  """Utility method to parse a JSON string. Raises HTTPError 502 on failure."""
+  try:
+    return json_loads(body)
+  except (ValueError, TypeError):
+    msg = 'Non-JSON response! Returning synthetic HTTP 502.\n%s' % body
+    logging.error(msg)
+    raise urllib.error.HTTPError(url, 502, msg, {}, None)
 
 
 def creation_result(content=None, description=None, abort=False,
                     error_plain=None, error_html=None):
-  """Create a new CreationResult named tuple, which the result of
-  create() and preview_create() to provides a detailed description of
-  publishing failures. If abort is False, we should continue looking
-  for an entry to publish; if True, we should immediately inform the
-  user. error_plain text is sent in response to failed publish
-  webmentions; error_html will be displayed to the user when
-  publishing interactively.
+  """Create a new :class:`CreationResult`.
+
+  :meth:`create()` and :meth:`preview_create()` use this to provide a detailed
+  description of publishing failures. If abort is False, we should continue
+  looking for an entry to publish; if True, we should immediately inform the
+  user. error_plain text is sent in response to failed publish webmentions;
+  error_html will be displayed to the user when publishing interactively.
 
   Args:
-    content: a string HTML snippet for preview_create() or a dict for create()
+
+    content: a string HTML snippet for :meth:`preview_create()` or a dict for
+      :meth:`create()`
     description: string HTML snippet describing the publish action, e.g.
       '@-reply' or 'RSVP yes to this event'. The verb itself is surrounded by a
       <span class="verb"> to allow styling. May also include <a> link(s) and
@@ -86,7 +134,7 @@ def creation_result(content=None, description=None, abort=False,
     error_html: a string
 
   Return:
-    a CreationResult named tuple
+    a :class:`CreationResult`
   """
   return CreationResult(content, description, abort, error_plain, error_html)
 
@@ -99,7 +147,8 @@ def object_type(obj):
   Args:
     obj: decoded JSON ActivityStreams object
 
-  Returns: string, ActivityStreams object type
+  Returns:
+    string: ActivityStreams object type
   """
   type = obj.get('objectType')
   return type if type and type != 'activity' else obj.get('verb')
@@ -111,44 +160,60 @@ class SourceMeta(type):
     cls = type.__new__(meta, name, bases, class_dict)
     name = getattr(cls, 'NAME', None)
     if name:
-      assert name.lower() not in sources, cls
       sources[name.lower()] = cls
     return cls
 
 
-class Source(object):
+class Source(object, metaclass=SourceMeta):
   """Abstract base class for a source (e.g. Facebook, Twitter).
 
   Concrete subclasses must override the class constants below and implement
-  get_activities().
+  :meth:`get_activities()`.
 
   Class constants:
-    DOMAIN: string, the source's domain
-    BASE_URL: optional, the source's base url
-    NAME: string, the source's human-readable name
-    FRONT_PAGE_TEMPLATE: string, the front page child template filename
-    AUTH_URL: string, the url for the "Authenticate" front page link
-    EMBED_POST: string, the HTML for embedding a post. Should have a %(url)s
-      placeholder for the post URL and (optionally) a %(content)s placeholder
-      for the post content.
+
+  * DOMAIN: string, the source's domain
+  * BASE_URL: optional, the source's base url
+  * NAME: string, the source's human-readable name
+  * FRONT_PAGE_TEMPLATE: string, the front page child template filename
+  * AUTH_URL: string, the url for the "Authenticate" front page link
+  * EMBED_POST: string, the HTML for embedding a post. Should have a %(url)s
+    placeholder for the post URL and (optionally) a %(content)s placeholder
+    for the post content.
+  * POST_ID_RE: regexp, optional, matches valid post ids. Used in post_id().
+  * HTML2TEXT_OPTIONS: dict mapping string html2text option names to values
+    https://github.com/Alir3z4/html2text/blob/master/docs/usage.md#available-options
+  * TRUNCATE_TEXT_LENGTH: integer, optional. Character limit to truncate to.
+        Defaults to Twitter's limit, 280 characters as of 2019-10-12.
+  * TRUNCATE_URL_LENGTH: integer, optional. Number of characters that URLs count
+        for. Defaults to Twitter's, 23 as of 2019-10-12.
   """
-  __metaclass__ = SourceMeta
+
+  POST_ID_RE = None
+  HTML2TEXT_OPTIONS = {}
+  TRUNCATE_TEXT_LENGTH = None
+  TRUNCATE_URL_LENGTH = None
 
   def user_url(self, user_id):
     """Returns the URL for a user's profile."""
     raise NotImplementedError()
 
   def get_actor(self, user_id=None):
-    """Returns the current user as a JSON ActivityStreams actor dict."""
+    """Fetches and returns a user.
+
+    Args: user_id: string, defaults to current user
+
+    Returns: dict, ActivityStreams actor object
+    """
     raise NotImplementedError()
 
   def get_activities(self, *args, **kwargs):
-    """Fetches and returns a list of ActivityStreams activities.
+    """Fetches and returns a list of activities.
 
     See get_activities_response() for args and kwargs.
 
     Returns:
-      list of ActivityStreams activity dicts
+      list, ActivityStreams activity dicts
     """
     return self.get_activities_response(*args, **kwargs)['items']
 
@@ -160,17 +225,17 @@ class Source(object):
                               fetch_mentions=False, search_query=None, **kwargs):
     """Fetches and returns ActivityStreams activities and response details.
 
-    Subclasses should override this. See get_activities() for an alternative
-    that just returns the list of activities.
+    Subclasses should override this. See :meth:`get_activities()` for an
+    alternative that just returns the list of activities.
 
     If user_id is provided, only that user's activity(s) are included.
     start_index and count determine paging, as described in the spec:
     http://activitystrea.ms/draft-spec.html#anchor14
 
-    app id is just object id
+    app id is just object id:
     http://opensocial-resources.googlecode.com/svn/spec/2.0/Social-Data.xml#appId
 
-    group id is string id of group or @self, @friends, @all, @search
+    group id is string id of group or @self, @friends, @all, @search:
     http://opensocial-resources.googlecode.com/svn/spec/2.0/Social-Data.xml#Group-ID
 
     The fetch_* kwargs all default to False because they often require extra API
@@ -181,7 +246,7 @@ class Source(object):
     Args:
       user_id: string, defaults to the currently authenticated user
       group_id: string, one of '@self', '@all', '@friends', '@search'. defaults
-        to 'friends'
+        to '@friends'
       app_id: string
       activity_id: string
       start_index: int >= 0
@@ -204,42 +269,48 @@ class Source(object):
       kwargs: some sources accept extra kwargs. See their docs for details.
 
     Returns:
-      response dict with values based on OpenSocial ActivityStreams REST API:
+      dict: response values based on OpenSocial ActivityStreams REST API.
         http://opensocial-resources.googlecode.com/svn/spec/2.0.1/Social-API-Server.xml#ActivityStreams-Service
         http://opensocial-resources.googlecode.com/svn/spec/2.0.1/Core-Data.xml
 
       It has these keys:
-        items: list of activity dicts
-        startIndex: int or None
-        itemsPerPage: int
-        totalResults: int or None (e.g. if it can 't be calculated efficiently)
-        filtered: False
-        sorted: False
-        updatedSince: False
-        etag: string etag returned by the API's initial call to get activities
+      * items: list of activity dicts
+      * startIndex: int or None
+      * itemsPerPage: int
+      * totalResults: int or None (e.g. if it can 't be calculated efficiently)
+      * filtered: False
+      * sorted: False
+      * updatedSince: False
+      * etag: string etag returned by the API's initial call to get activities
+
+    Raises:
+      :class:`ValueError`: if any argument is invalid for this source
+      :class:`NotImplementedError`: if the source doesn't support the requested
+        operation, e.g. Facebook doesn't support search.
     """
     raise NotImplementedError()
 
   @classmethod
   def make_activities_base_response(cls, activities, *args, **kwargs):
-    """Generates a base response dict for get_activities_response().
+    """Generates a base response dict for :meth:`get_activities_response()`.
 
-    See get_activities() for args and kwargs.
+    See :meth:`get_activities()` for args and kwargs.
     """
-    return {'startIndex': kwargs.get('start_index', 0),
-            'itemsPerPage': len(activities),
-            'totalResults': None if kwargs.get('activity_id') else len(activities),
-            # TODO: this is just for compatibility with
-            # http://activitystreamstester.appspot.com/
-            # the OpenSocial spec says to use entry instead, so switch back
-            # to that eventually
-            'items': activities,
-            'filtered': False,
-            'sorted': False,
-            'updatedSince': False,
-            }
+    return {
+      'startIndex': kwargs.get('start_index', 0),
+      'itemsPerPage': len(activities),
+      'totalResults': None if kwargs.get('activity_id') else len(activities),
+      # TODO: this is just for compatibility with
+      # http://activitystreamstester.appspot.com/
+      # the OpenSocial spec says to use entry instead, so switch back
+      # to that eventually
+      'items': activities,
+      'filtered': False,
+      'sorted': False,
+      'updatedSince': False,
+    }
 
-  def create(self, obj, include_link=False, ignore_formatting=False):
+  def create(self, obj, include_link=OMIT_LINK, ignore_formatting=False):
     """Creates a new object: a post, comment, like, share, or RSVP.
 
     Subclasses should override this. Different sites will support different
@@ -249,24 +320,23 @@ class Source(object):
     Args:
       obj: ActivityStreams object. At minimum, must have the content field.
         objectType is strongly recommended.
-      include_link: boolean. If True, includes a link to the object
-        (if it has one) in the content.
+      include_link: string. 'include', 'omit', or 'if truncated'; whether to
+        include a link to the object (if it has one) in the content.
       ignore_formatting: whether to use content text as is, instead of
         converting its HTML to plain text styling (newlines, etc.)
 
     Returns:
-      a CreationResult, whose contents will be a dict
-
-      The dict may be None or empty. If the newly created
-      object has an id or permalink, they'll be provided in the values
-      for 'id' and 'url'.
+      CreationResult: contents will be a dict. The dict may be None or empty. If
+        the newly created object has an id or permalink, they'll be provided in
+        the values for 'id' and 'url'.
     """
     raise NotImplementedError()
 
-  def preview_create(self, obj, include_link=False, ignore_formatting=False):
+  def preview_create(self, obj, include_link=OMIT_LINK, ignore_formatting=False):
     """Previews creating a new object: a post, comment, like, share, or RSVP.
 
-    Returns HTML that previews what create() with the same object will do.
+    Returns HTML that previews what :meth:`create()` with the same object will
+    do.
 
     Subclasses should override this. Different sites will support different
     functionality; check each subclass for details. The actor will usually be
@@ -275,29 +345,54 @@ class Source(object):
     Args:
       obj: ActivityStreams object. At minimum, must have the content field.
         objectType is strongly recommended.
-      include_link: boolean. If True, includes a link to the object
+      include_link: string. Whether to include a link to the object
         (if it has one) in the content.
       ignore_formatting: whether to use content text as is, instead of
         converting its HTML to plain text styling (newlines, etc.)
 
     Returns:
-      a CreationResult, whose contents will be a unicode string
-      HTML snippet (or None)
+      CreationResult: contents will be a unicode string HTML snippet (or None)
+    """
+    raise NotImplementedError()
+
+  def delete(self, id):
+    """Deletes a post.
+
+    Generally only supports posts that were authored by the authenticating user.
+
+    Args:
+      id: silo object id
+
+    Returns:
+      CreationResult
+    """
+    raise NotImplementedError()
+
+  def preview_delete(self, id):
+    """Previews deleting a post.
+
+    Args:
+      id: silo object id
+
+    Returns:
+      CreationResult
     """
     raise NotImplementedError()
 
   def get_event(self, event_id):
-    """Returns a ActivityStreams event activity.
+    """Fetches and returns an event.
 
     Args:
       id: string, site-specific event id
 
-    Returns: dict, decoded ActivityStreams activity, or None
+    Returns:
+      dict: decoded ActivityStreams activity, or None
     """
     raise NotImplementedError()
 
-  def get_comment(self, comment_id, activity_id=None, activity_author_id=None):
-    """Returns an ActivityStreams comment object.
+  def get_comment(self, comment_id, activity_id=None, activity_author_id=None,
+                  activity=None):
+    """Fetches and returns a comment.
 
     Subclasses should override this.
 
@@ -306,11 +401,17 @@ class Source(object):
       activity_id: string activity id, optional
       activity_author_id: string activity author id, optional. Needed for some
         sources (e.g. Facebook) to construct the comment permalink.
+      activity: activity object, optional. May avoid an API call if provided.
+
+    Returns: dict, ActivityStreams comment object
+
+    Raises:
+      :class:`ValueError`: if any argument is invalid for this source
     """
     raise NotImplementedError()
 
-  def get_like(self, activity_user_id, activity_id, like_user_id):
-    """Returns an ActivityStreams 'like' activity object.
+  def get_like(self, activity_user_id, activity_id, like_user_id, activity=None):
+    """Fetches and returns a 'like'.
 
     Default implementation that fetches the activity and its likes, then
     searches for a like by the given user. Subclasses should override this if
@@ -320,32 +421,89 @@ class Source(object):
       activity_user_id: string id of the user who posted the original activity
       activity_id: string activity id
       like_user_id: string id of the user who liked the activity
-    """
-    activities = self.get_activities(user_id=activity_user_id,
-                                     activity_id=activity_id,
-                                     fetch_likes=True)
-    return self._get_tag(activities, 'like', like_user_id)
+      activity: activity object, optional. May avoid an API call if provided.
 
-  def get_share(self, activity_user_id, activity_id, share_id):
-    """Returns an ActivityStreams 'share' activity object.
+    Returns: dict, ActivityStreams 'like' activity object
+    """
+    if not activity:
+      activity = self._get_activity(activity_user_id, activity_id, fetch_likes=True)
+    return self._get_tag(activity, 'like', like_user_id)
+
+  def get_reaction(self, activity_user_id, activity_id, reaction_user_id,
+                   reaction_id, activity=None):
+    """Fetches and returns a 'reaction'.
+
+    Default implementation that fetches the activity and its reactions, then
+    searches for this specific reaction. Subclasses should override this if they
+    can optimize the process.
+
+    Args:
+      activity_user_id: string id of the user who posted the original activity
+      activity_id: string activity id
+      reaction_user_id: string id of the user who reacted
+      reaction_id: string id of the reaction
+      activity: activity object, optional. May avoid an API call if provided.
+
+    Returns: dict, ActivityStreams 'reaction' activity object
+    """
+    if not activity:
+      activity = self._get_activity(activity_user_id, activity_id, fetch_likes=True)
+    return self._get_tag(activity, 'react', reaction_user_id, reaction_id)
+
+  def get_share(self, activity_user_id, activity_id, share_id, activity=None):
+    """Fetches and returns a 'share'.
 
     Args:
       activity_user_id: string id of the user who posted the original activity
       activity_id: string activity id
       share_id: string id of the share object or the user who shared it
-    """
-    activities = self.get_activities(user_id=activity_user_id,
-                                     activity_id=activity_id,
-                                     fetch_shares=True)
-    return self._get_tag(activities, 'share', share_id)
+      activity: activity object, optional. May avoid an API call if provided.
 
-  def get_rsvp(self, activity_user_id, event_id, user_id):
-    """Returns an ActivityStreams 'rsvp-*' activity object.
+    Returns: dict, an ActivityStreams 'share' activity object
+    """
+    if not activity:
+      activity = self._get_activity(activity_user_id, activity_id, fetch_shares=True)
+    return self._get_tag(activity, 'share', share_id)
+
+  def get_rsvp(self, activity_user_id, event_id, user_id, event=None):
+    """Fetches and returns an RSVP.
 
     Args:
-      activity_user_id: string id of the user who posted the original activity
+      activity_user_id: string id of the user who posted the event. unused.
       event_id: string event id
-      user_id: string id of the user who RSVPed
+      user_id: string user id
+      event: AS event activity (optional)
+
+    Returns: dict, an ActivityStreams RSVP activity object
+    """
+    if not event:
+      event = self.get_event(event_id)
+      if not event:
+        return None
+
+    for rsvp in self.get_rsvps_from_event(event['object']):
+      for field in 'actor', 'object':
+        id = rsvp.get(field, {}).get('id')
+        if id and user_id == util.parse_tag_uri(id)[1]:
+          return rsvp
+
+  def get_blocklist(self):
+    """Fetches and returns the current user's block list.
+
+    ...ie the users that the current user is blocking. The exact semantics of
+    "blocking" vary from silo to silo.
+
+    Returns:
+      sequence of actor objects
+    """
+    raise NotImplementedError()
+
+  def get_blocklist_ids(self):
+    """Returns the current user's block list as a list of silo-specific user ids.
+
+    Returns:
+      sequence of user ids, not globally unique across other sources. May be
+        integers or strings.
     """
     raise NotImplementedError()
 
@@ -357,49 +515,36 @@ class Source(object):
     dicts with 'value': URL.
 
     Args:
-      user: dict, a decoded JSON Facebook user
+      user: dict, a decoded JSON silo user object
 
     Returns:
-      an ActivityStreams actor dict, ready to be JSON-encoded
+      dict: ActivityStreams actor, ready to be JSON-encoded
     """
     raise NotImplementedError()
 
-  def _get_tag(self, activities, verb, user_id):
-    if not activities:
+  def _get_activity(self, user_id, activity_id, **kwargs):
+    activities = self.get_activities(user_id=user_id, activity_id=activity_id,
+                                     **kwargs)
+    if activities:
+      return activities[0]
+
+  def _get_tag(self, activity, verb, user_id, tag_id=None):
+    if not activity:
       return None
 
     user_tag_id = self.tag_uri(user_id)
-    for tag in activities[0].get('object', {}).get('tags', []):
+    if tag_id:
+      tag_id = self.tag_uri(tag_id)
+
+    for tag in activity.get('object', {}).get('tags', []):
       author = tag.get('author', {})
       if (tag.get('verb') == verb and
+          (not tag_id or tag_id == tag.get('id')) and
           (author.get('id') == user_tag_id or author.get('numeric_id') == user_id)):
         return tag
 
-  def _fetch_like(self, activity):
-    """Fetches and injects likes into an activity, in place.
-
-    Subclasses should usually override.
-
-    Args:
-      activities: ActivityStreams activity dict
-
-    Returns: same activity dict
-    """
-    return activity
-
-  def _fetch_share(self, activity):
-    """Fetches and injects shares into an activity, in place.
-
-    Subclasses should usually override.
-
-    Args:
-      activities: ActivityStreams activity dict
-
-    Returns: same activity dict
-    """
-    return activity
-
-  def postprocess_activity(self, activity):
+  @staticmethod
+  def postprocess_activity(activity):
     """Does source-independent post-processing of an activity, in place.
 
     Right now just populates the title field.
@@ -421,28 +566,31 @@ class Source(object):
       'share': 'shared',
     }
 
-    actor_name = self.actor_name(activity.get('actor'))
+    actor_name = Source.actor_name(activity.get('actor'))
     obj = activity.get('object')
 
-    if obj and not activity.get('title'):
-      verb = DISPLAY_VERBS.get(activity['verb'])
-      obj_name = obj.get('displayName')
-      obj_type = TYPE_DISPLAY_NAMES.get(obj.get('objectType'))
-      if obj_name and not verb:
-        activity['title'] = obj_name
-      elif verb and (obj_name or obj_type):
-        app = activity.get('generator', {}).get('displayName')
-        name = obj_name if obj_name else 'a %s' % (obj_type or 'unknown')
-        app = ' on %s' % app if app else ''
-        activity['title'] = '%s %s %s%s.' % (actor_name, verb or 'posted',
-                                             name, app)
+    if obj:
+      activity['object'] = Source.postprocess_object(obj)
+      if not activity.get('title'):
+        verb = DISPLAY_VERBS.get(activity.get('verb'))
+        obj_name = obj.get('displayName')
+        obj_type = TYPE_DISPLAY_NAMES.get(obj.get('objectType'))
+        if obj_name and not verb:
+          activity['title'] = obj_name
+        elif verb and (obj_name or obj_type):
+          app = activity.get('generator', {}).get('displayName')
+          name = obj_name if obj_name else 'a %s' % (obj_type or 'unknown')
+          app = ' on %s' % app if app else ''
+          activity['title'] = '%s %s %s%s.' % (actor_name, verb or 'posted',
+                                               name, app)
 
     return util.trim_nulls(activity)
 
-  def postprocess_object(self, obj):
+  @staticmethod
+  def postprocess_object(obj):
     """Does source-independent post-processing of an object, in place.
 
-    * populates location.position based on latitude and longitude
+    Populates location.position based on latitude and longitude.
 
     Args:
       object: object dict
@@ -479,20 +627,20 @@ class Source(object):
       domains: optional sequence of domains. If provided, only links to these
         domains will be considered original and stored in upstreamDuplicates.
         (Permashortcitations are exempt.)
-      cache: optional, a cache object for storing resolved URL redirects. Passed
-        to follow_redirects().
       include_redirect_sources: boolean, whether to include URLs that redirect
         as well as their final destination URLs
+      cache: deprecated, unused
       kwargs: passed to requests.head() when following redirects
 
-    Returns: ([string original post URLs], [string mention URLs]) tuple
+    Returns:
+      ([string original post URLs], [string mention URLs]) tuple
     """
     obj = activity.get('object') or activity
     content = obj.get('content', '').strip()
 
     # find all candidate URLs
     tags = [t.get('url') for t in obj.get('attachments', []) + obj.get('tags', [])
-            if t.get('objectType') in ('article', 'mention', None)]
+            if t.get('objectType') in ('article', 'mention', 'note', None)]
     candidates = tags + util.extract_links(content) + obj.get('upstreamDuplicates', [])
 
     # Permashortcitations (http://indiewebcamp.com/permashortcitation) are short
@@ -501,19 +649,20 @@ class Source(object):
     candidates += [match.expand(r'http://\1/\2') for match in
                    Source._PERMASHORTCITATION_RE.finditer(content)]
 
-    candidates = set(filter(None,
-      (util.clean_url(url) for url in candidates
-       # heuristic: ellipsized URLs are probably incomplete, so omit them.
-       if url and not url.endswith('...') and not url.endswith(u'…'))))
+    candidates = set(util.dedupe_urls(
+      util.clean_url(url) for url in candidates
+      # heuristic: ellipsized URLs are probably incomplete, so omit them.
+      if url and not url.endswith('...') and not url.endswith('…')))
 
     # check for redirect and add their final urls
     redirects = {}  # maps final URL to original URL for redirects
-    for url in list(candidates):
-      resolved = util.follow_redirects(url, cache=cache, **kwargs)
+    for url in candidates:
+      resolved = util.follow_redirects(url, **kwargs)
       if (resolved.url != url and
           resolved.headers.get('content-type', '').startswith('text/html')):
         redirects[resolved.url] = url
-        candidates.add(resolved.url)
+
+    candidates.update(redirects.keys())
 
     # use domains to determine which URLs are original post links vs mentions
     originals = set()
@@ -546,11 +695,11 @@ class Source(object):
   def is_public(obj):
     """Returns True if the object is public, False if private, None if unknown.
 
-    ...according to the Audience Targeting extension
-    https://developers.google.com/+/api/latest/activities/list#collection
+    ...according to the Audience Targeting extension:
+    http://activitystrea.ms/specs/json/targeting/1.0/
 
-    Expects values generated by this library: objectType group, alias @public or
-    @private.
+    Expects values generated by this library: objectType group, alias @public,
+    @unlisted, or @private.
 
     Also, important point: this defaults to true, ie public. Bridgy depends on
     that and prunes the to field from stored activities in Response objects (in
@@ -560,33 +709,34 @@ class Source(object):
     to = obj.get('to') or obj.get('object', {}).get('to') or []
     aliases = util.trim_nulls([t.get('alias') for t in to])
     object_types = util.trim_nulls([t.get('objectType') for t in to])
-    return (True if '@public' in aliases
+    return (True if '@public' in aliases or '@unlisted' in aliases
             else None if 'unknown' in object_types
             else False if aliases
             else True)
 
   @staticmethod
   def add_rsvps_to_event(event, rsvps):
-    """Adds RSVP objects to an event's *attending fields, in place.
+    """Adds RSVP objects to an event's attending fields, in place.
 
     Args:
       event: ActivityStreams event object
       rsvps: sequence of ActivityStreams RSVP activity objects
     """
     for rsvp in rsvps:
-      field = RSVP_TO_EVENT.get(rsvp.get('verb'))
+      field = RSVP_VERB_TO_COLLECTION.get(rsvp.get('verb'))
       if field:
         event.setdefault(field, []).append(rsvp.get(
             'object' if field == 'invited' else 'actor'))
 
   @staticmethod
   def get_rsvps_from_event(event):
-    """Returns RSVP objects for an event's *attending fields.
+    """Returns RSVP objects for an event's attending fields.
 
     Args:
       event: ActivityStreams event object
 
-    Returns: sequence of ActivityStreams RSVP activity objects
+    Returns:
+      sequence of ActivityStreams RSVP activity objects
     """
     id = event.get('id')
     if not id:
@@ -596,20 +746,26 @@ class Source(object):
       return []
     domain, event_id = parsed
     url = event.get('url')
+    author = event.get('author')
 
     rsvps = []
-    for verb, field in RSVP_TO_EVENT.items():
+    for verb, field in RSVP_VERB_TO_COLLECTION.items():
       for actor in event.get(field, []):
         rsvp = {'objectType': 'activity',
                 'verb': verb,
-                'actor': actor,
+                'object' if verb == 'invite' else 'actor': actor,
                 'url': url,
                 }
+
         if event_id and 'id' in actor:
           _, actor_id = util.parse_tag_uri(actor['id'])
           rsvp['id'] = util.tag_uri(domain, '%s_rsvp_%s' % (event_id, actor_id))
           if url:
             rsvp['url'] = '#'.join((url, actor_id))
+
+        if verb == 'invite' and author:
+          rsvp['actor'] = author
+
         rsvps.append(rsvp)
 
     return rsvps
@@ -629,8 +785,8 @@ class Source(object):
     Args:
       before, after: dicts, ActivityStreams activities or objects
 
-    Returns: boolean
-
+    Returns:
+      boolean
     """
     def changed(b, a, field, label):
       b_val = b.get(field)
@@ -649,15 +805,32 @@ class Source(object):
 
   @classmethod
   def embed_post(cls, obj):
-    """Returns the HTML string for embedding a post object."""
+    """Returns the HTML string for embedding a post object.
+
+    Args:
+      obj: AS1 dict with at least url, and optionally also content.
+
+    Returns: string, HTML
+
+    Raises: ValueError, if obj['content'] contains raw < or > characters.
+    """
     obj = copy.copy(obj)
     for field in 'url', 'content':
       if field not in obj:
         obj.setdefault(field, obj.get('object', {}).get(field, ''))
+
+    if '<' in obj['content'] or '>' in obj['content']:
+      raise ValueError("obj['content'] has unescaped < or > characters!")
+
+    # escape URL, but not with urllib.parse.quote, because it quotes a ton of
+    # chars we want to pass through, including most unicode chars.
+    obj['url'] = obj['url'].replace('<', '%3C').replace('>', '%3E')
+
     return cls.EMBED_POST % obj
 
   @classmethod
   def embed_actor(cls, actor):
+    """Returns the HTML string for embedding an actor object."""
     return """
 <a class="h-card" href="%s">
  <img class="profile u-photo" src="%s" width="32px" /> %s</a>""" % (
@@ -681,7 +854,8 @@ class Source(object):
     Args:
       obj: ActivityStreams object
 
-    Returns: dict, minimal ActivityStreams object. Usually has at least id; may
+    Returns:
+      dict, minimal ActivityStreams object. Usually has at least id; may
       also have url, author, etc.
     """
     # look at in-reply-tos first, then objects (for likes and reposts).
@@ -689,12 +863,8 @@ class Source(object):
     # singular, but microformats2.json_to_object() sometimes returns activities
     # that have a list value, e.g. likes or reposts of multiple objects.
     candidates = []
-    for field in ('inReplyTo', 'object'):
-      objs = obj.get(field, [])
-      if isinstance(objs, dict):
-        candidates.append(objs)
-      else:
-        candidates += objs
+    for field in ('inReplyTo', 'object', 'target'):
+      candidates += util.get_list(obj, field)
 
     for base_obj in candidates:
       parsed_id = util.parse_tag_uri(base_obj.get('id', ''))
@@ -707,30 +877,46 @@ class Source(object):
     else:
       return {}
 
-    base_obj = copy.deepcopy(base_obj)
-    id = base_obj.get('id')
-    url = base_obj.get('url')
+    return self._postprocess_base_object(base_obj)
+
+  @classmethod
+  def _postprocess_base_object(cls, obj):
+    obj = copy.deepcopy(obj)
+    id = obj.get('id')
+    url = obj.get('url')
 
     if id:
       parsed = util.parse_tag_uri(id)
       if parsed:
-        base_obj['id'] = parsed[1]
+        obj['id'] = parsed[1]
     elif url:
-      base_obj['id'] = self.post_id(url)
+      obj['id'] = cls.base_id(url)
 
-    return base_obj
+    return obj
+
+  @classmethod
+  def base_id(cls, url):
+    """Guesses the id of the object in the given URL.
+
+    Returns:
+      string, or None
+    """
+    return urllib.parse.urlparse(url).path.rstrip('/').rsplit('/', 1)[-1] or None
 
   @classmethod
   def post_id(cls, url):
     """Guesses the post id of the given URL.
 
-    Returns: string, or None
+    Returns:
+      string, or None
     """
-    return urlparse.urlparse(url).path.rstrip('/').rsplit('/', 1)[-1] or None
+    id = cls.base_id(url)
+    if id and (not cls.POST_ID_RE or cls.POST_ID_RE.match(id)):
+      return id
 
   def _content_for_create(self, obj, ignore_formatting=False, prefer_name=False,
-                          strip_first_video_tag=False):
-    """Returns the content text to use in create() and preview_create().
+                          strip_first_video_tag=False, strip_quotations=False):
+    """Returns content text for :meth:`create()` and :meth:`preview_create()`.
 
     Returns summary if available, then content, then displayName.
 
@@ -744,27 +930,99 @@ class Source(object):
       prefer_name: boolean, whether to prefer displayName to content
       strip_first_video_tag: if true, removes the first <video> tag. useful when
         it will be uploaded and attached to the post natively in the silo.
+      strip_quotations: if true, removes .u-quotation-of tags. useful when
+        creating quote tweets.
 
-    Returns: string, possibly empty
+    Returns:
+      string, possibly empty
     """
     summary = obj.get('summary', '').strip()
     name = obj.get('displayName', '').strip()
     content = obj.get('content', '').strip()
 
+    # note that unicode() on a BeautifulSoup object preserves HTML and
+    # whitespace, even after modifying the DOM, which is important for
+    # formatting.
+    #
+    # The catch is that it adds a '<html><head></head><body>' header and
+    # '</body></html>' footer. ah well. harmless.
+    soup = util.parse_html(content)
     if strip_first_video_tag:
-      # don't use BeautifulSoup because we want to preserve exact formatting
-      # and whitespace otherwise
-      content = re.sub(r'<video[^<>]*>.*</video>', '', content, count=1)
+      video = soup.video or soup.find(class_='u-video')
+      if video:
+        video.extract()
+        content = str(soup)
 
-    if summary == strip_html_tags(content).strip():
+    if strip_quotations:
+      quotations = soup.find_all(class_='u-quotation-of')
+      if quotations:
+        for q in quotations:
+          q.extract()
+        content = str(soup)
+
+    # compare to content with HTML tags stripped
+    if summary == soup.get_text('').strip():
       # summary and content are the same; prefer content so that we can use its
       # HTML formatting.
       summary = None
 
-    if not ignore_formatting:
-      content = html_to_text(content)
+    # sniff whether content is HTML or plain text. use html.parser instead of
+    # the default html5lib since html.parser is stricter and expects actual
+    # HTML tags.
+    # https://www.crummy.com/software/BeautifulSoup/bs4/doc/#differences-between-parsers
+    is_html = (bool(BeautifulSoup(content, 'html.parser').find()) or
+               HTML_ENTITY_RE.search(content))
+    if is_html and not ignore_formatting:
+      content = html_to_text(content, **self.HTML2TEXT_OPTIONS)
+    elif not is_html and ignore_formatting:
+      content = re.sub(r'\s+', ' ', content)
 
     return summary or (
             (name or content) if prefer_name else
             (content or name)
-           ) or u''
+           ) or ''
+
+  def truncate(self, content, url, include_link, type=None, quote_url=None):
+    """Shorten text content to fit within a character limit.
+
+    Character limit and URL character length are taken from the
+    TRUNCATE_TEXT_LENGTH and TRUNCATE_URL_LENGTH class constants
+
+    Args:
+      content: string
+      url: string
+      include_link: string
+      type: string, optional: 'article', 'note', etc.
+      quote_url: string URL, optional. If provided, it will be appended to the
+        content, *after* truncating.
+
+    Return: string, the possibly shortened and ellipsized text
+    """
+    kwargs = {}
+
+    if quote_url:
+      kwargs['target_length'] = (
+        (self.TRUNCATE_TEXT_LENGTH or brevity.WEIGHTS['maxWeightedTweetLength']) -
+        (self.TRUNCATE_URL_LENGTH or brevity.WEIGHTS['transformedURLLength']) - 1)
+    elif self.TRUNCATE_TEXT_LENGTH is not None:
+      kwargs['target_length'] = self.TRUNCATE_TEXT_LENGTH
+
+    if self.TRUNCATE_URL_LENGTH is not None:
+      kwargs['link_length'] = self.TRUNCATE_URL_LENGTH
+
+    if include_link != OMIT_LINK:
+      kwargs['permalink'] = url  # only include when text is truncated
+
+    if include_link == INCLUDE_LINK:
+      kwargs['permashortlink'] = url  # always include
+
+    if type == 'article':
+      kwargs['format'] = brevity.FORMAT_ARTICLE
+
+    truncated = brevity.shorten(content, **kwargs)
+
+    if quote_url:
+      truncated += ' ' + quote_url
+
+    return truncated
+
